from starlette.requests import Request

from throttled import Hit
from throttled.fastapi.limiters.base import Limiter, Middleware


class HostBasedLimiter(Limiter, Middleware):
    def _maybe_block_request(self, request: Request):
        self(request)

    def __call__(self, request: Request):  # pylint: disable=arguments-differ
        hit = Hit(key=request.client.host)
        self._strategy(hit)



from typing import Optional


class RateLimitExceeded(Exception):
    def __init__(self, key: str, retry_after: Optional[int] = None):
        self.key = key
        self.retry_after = retry_after
        super().__init__()



from collections import deque
from typing import Deque, MutableMapping, NamedTuple, Optional

from throttled import Hit, Rate
from throttled.exceptions import RateLimitExceeded
from throttled.strategy.abstract import Strategy


class HitsWindow(NamedTuple):
    initialized_at: float
    hits: Deque[Hit]


class FixedWindowStrategy(Strategy):
    def __init__(
        self, limit: Rate, cache: Optional[MutableMapping[str, HitsWindow]] = None
    ):
        self.__limit = limit
        self.__cache: MutableMapping[str, HitsWindow] = cache or {}

    def __get_time_spent_in_window(self, hit: Hit) -> float:
        return hit.time % self.__limit.interval

    def __maybe_reset(self, hit: Hit):
        initialized_at = hit.time - self.__get_time_spent_in_window(hit)

        if hit.key not in self.__cache:
            self.__cache[hit.key] = HitsWindow(initialized_at, deque())
            return

        window = self.__cache[hit.key]
        if initialized_at > window.initialized_at:
            self.__cache[hit.key] = HitsWindow(initialized_at, deque())

    def __call__(self, hit: Hit):
        self.__maybe_reset(hit)
        rate = Rate.from_hits(self.__cache[hit.key].hits)
        if rate > self.__limit:
            raise RateLimitExceeded(
                hit.key,
                retry_after=int(
                    self.__limit.interval - self.__get_time_spent_in_window(hit)
                ),
            )
        self.__cache[hit.key].hits.append(hit)



import itertools
import sys
import time as clock
from dataclasses import dataclass, field
from functools import reduce
from typing import Iterable, Union

DATACLASS_KWARGS = {"frozen": True}
if sys.version_info >= (3, 10):
    DATACLASS_KWARGS["slots"] = True


@dataclass(**DATACLASS_KWARGS)
class Hit:
    key: str = field(
        default_factory=str,
        metadata={"description": "The base contribution to the hit key."},
    )
    time: float = field(
        default_factory=clock.time,
        metadata={
            "units": "seconds",
            "description": "time the hit occurred since epoch.",
        },
    )
    cost: int = 1

    def __lt__(self, other: "Hit") -> bool:
        return self.time < other.time

    def __gt__(self, other: "Hit") -> bool:
        return self.time > other.time


@dataclass(**DATACLASS_KWARGS)
class Rate:
    """
    :param hits: number of hits
    :param interval: interval for the hits in seconds
    :return: a Rate object
    """

    hits: int
    interval: float

    @property
    def ratio(self) -> float:
        """Return the rate as float in hits/seconds units."""
        return max(self.hits, 0) / max(self.interval, 1e-2)

    @classmethod
    def __assert_class(cls, other) -> "Rate":
        if not isinstance(other, Rate):
            raise TypeError(
                f"Can't compare {cls.__name__} with {other.__class__.__name__}"
            )
        return other

    def __lt__(self, other: Union["Rate", int, float]) -> bool:
        if isinstance(other, (int, float)):
            return self.ratio < other
        return self.ratio < other.ratio

    def __gt__(self, other: Union["Rate", int, float]) -> bool:
        if isinstance(other, (int, float)):
            return self.ratio > other
        return self.ratio > other.ratio

    def __eq__(self, other) -> bool:
        if isinstance(other, (int, float)):
            return self.ratio == other
        other = self.__assert_class(other)
        return self.ratio == other.ratio

    @classmethod
    def from_hits(cls, hits: Iterable[Hit]) -> "Rate":
        it1, it2, it3 = itertools.tee(hits, 3)
        cost: int = reduce(lambda a, b: a + b.cost, it1, 0)
        return cls(
            hits=cost,
            interval=max(it2).time - min(it3).time if cost > 0 else 0,
        )



from typing import List, Sequence, Union

from fastapi import Depends, FastAPI
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction

from .limiters.base import Limiter, Middleware, catcher_middleware


class APILimiter:
    def __init__(self, use_middleware_where_possible: bool = True):
        self.__dependency_limiters: List[Limiter] = []
        self.__middleware_limiters: List[Middleware] = []
        self.__use_middleware_where_possible = use_middleware_where_possible

    def append(self, limiter: Union[Limiter, Middleware]):
        if isinstance(limiter, Middleware) and self.__use_middleware_where_possible:
            self.__middleware_limiters.append(limiter)
            return
        if isinstance(limiter, Limiter):
            self.__dependency_limiters.append(limiter)
            return
        raise TypeError(f"Object {limiter} is not a Limiter.")

    @property
    def dependencies(self) -> Sequence[Depends]:
        return tuple(Depends(limiter) for limiter in self.__dependency_limiters)

    @property
    def middlewares(self) -> Sequence[DispatchFunction]:
        funcs = [catcher_middleware]
        for limiter in self.__middleware_limiters:
            funcs.append(limiter.dispatch)
        return funcs

    def add_middleware(self, app: FastAPI):
        for middleware in self.middlewares:
            app.add_middleware(BaseHTTPMiddleware, dispatch=middleware)



from starlette.requests import Request

from throttled import Hit
from throttled.fastapi.limiters.base import Limiter, Middleware


class TotalLimiter(Limiter, Middleware):
    def _maybe_block_request(self, request: Request):
        self()

    def __call__(self):  # pylint: disable=arguments-differ
        self._strategy(Hit())



from abc import ABC, abstractmethod

from throttled import Hit


class Strategy(ABC):
    @abstractmethod
    def __call__(self, hit: Hit):
        """
        :param hit: The hit to be tested
        :raises: RateLimitExceeded
        """



from abc import ABC, abstractmethod
from typing import Sequence

from starlette import status
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from throttled.exceptions import RateLimitExceeded
from throttled.strategy import Strategy


def response_from_exception(exc: RateLimitExceeded) -> Response:
    return Response(
        status_code=status.HTTP_429_TOO_MANY_REQUESTS,
        content=f"Rate exceeded for key={exc.key}.",
        headers={"Retry-After": str(exc.retry_after)},
    )


async def catcher_middleware(request: Request, call_next):
    try:
        return await call_next(request)
    except RateLimitExceeded as exc:
        return response_from_exception(exc)


class Middleware(ABC):
    _ignored_paths: Sequence[str] = ("docs", "redoc", "favicon.ico", "openapi.json")

    def ignore_path(self, path: str):
        ignored = list(self._ignored_paths)
        ignored.append(path)
        self._ignored_paths = ignored

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            path = str(request.url).replace(str(request.base_url), "")
            if path not in self._ignored_paths:
                self._maybe_block_request(request)
            return await call_next(request)
        except RateLimitExceeded as exc:
            return response_from_exception(exc)

    @abstractmethod
    def _maybe_block_request(self, request: Request):
        ...


class Limiter(ABC):
    def __init__(self, strategy: Strategy):
        self._strategy = strategy

    def __call__(self, *args, **kwargs):
        ...
